@@ -36,6 +36,7 @@
 #include <QDir>
 #include <QDirIterator>
 #include <QMenuBar>
+#include <QMetaType>
 #include <QTime>
 #include <QTimer>
 #include <QDesktopWidget>
@@ -88,6 +89,10 @@ JuK::JuK(const QStringList &filesToOpen, QWidget *parent) :
 {
     // Expect segfaults if you change this order.
 
+    // Allow to be passed across threads
+    qRegisterMetaType<FileHandle>();
+    qRegisterMetaType<FileHandleList>();
+
     m_instance = this;
 
     readSettings();
@@ -138,7 +143,6 @@ JuK::JuK(const QStringList &filesToOpen, QWidget *parent) :
                                QStringLiteral("playing audio"));
             if (reply.isValid()) {
                 m_pmToken = reply.value();
-                qCDebug(JUK_LOG) << "Inhibiting power management";
             }
         }
     });
@@ -149,7 +153,6 @@ JuK::JuK(const QStringList &filesToOpen, QWidget *parent) :
             reply = pmInterface->call(QStringLiteral("UnInhibit"), m_pmToken);
             if (reply.errorName().isEmpty()) {
                 m_pmToken = 0;
-                qCDebug(JUK_LOG) << "Uninhibiting power management";
             }
         }
     };
