@@ -55,12 +55,16 @@
 #include <QScrollBar>
 #include <QPainter>
 
+#include <QtConcurrent>
+#include <QFutureWatcher>
+
 #include <id3v1genres.h>
 
 #include <time.h>
 #include <cmath>
 #include <algorithm>
 
+#include "directoryloader.h"
 #include "playlistitem.h"
 #include "playlistcollection.h"
 #include "playlistsearch.h"
@@ -111,94 +115,55 @@ int                     Playlist::m_leftColumn     = 0;
 // public members
 ////////////////////////////////////////////////////////////////////////////////
 
-Playlist::Playlist(PlaylistCollection *collection, const QString &name,
-                   const QString &iconName) :
-    QTreeWidget(collection->playlistStack()),
-    m_collection(collection),
-    m_fetcher(new WebImageFetcher(this)),
-    m_allowDuplicates(true),
-    m_applySharedSettings(true),
-    m_columnWidthModeChanged(false),
-    m_disableColumnWidthUpdates(true),
-    m_time(0),
-    m_widthsDirty(true),
-    m_searchEnabled(true),
-    m_playlistName(name),
-    m_rmbMenu(0),
-    m_toolTip(0),
-    m_blockDataChanged(false)
+Playlist::Playlist(
+        bool delaySetup, const QString &name,
+        PlaylistCollection *collection, const QString &iconName,
+        int extraCols)
+    : QTreeWidget(collection->playlistStack())
+    , m_collection(collection)
+    , m_playlistName(name)
+    , m_fetcher(new WebImageFetcher(this))
 {
-    setup();
-    collection->setupPlaylist(this, iconName);
-}
-
-Playlist::Playlist(PlaylistCollection *collection, const PlaylistItemList &items,
-                   const QString &name, const QString &iconName) :
-    QTreeWidget(collection->playlistStack()),
-    m_collection(collection),
-    m_fetcher(new WebImageFetcher(this)),
-    m_allowDuplicates(true),
-    m_applySharedSettings(true),
-    m_columnWidthModeChanged(false),
-    m_disableColumnWidthUpdates(true),
-    m_time(0),
-    m_widthsDirty(true),
-    m_searchEnabled(true),
-    m_playlistName(name),
-    m_rmbMenu(0),
-    m_toolTip(0),
-    m_blockDataChanged(false)
-{
-    setup();
-    collection->setupPlaylist(this, iconName);
-    createItems(items);
-}
-
-Playlist::Playlist(PlaylistCollection *collection, const QFileInfo &playlistFile,
-                   const QString &iconName) :
-    QTreeWidget(collection->playlistStack()),
-    m_collection(collection),
-    m_fetcher(new WebImageFetcher(this)),
-    m_allowDuplicates(true),
-    m_applySharedSettings(true),
-    m_columnWidthModeChanged(false),
-    m_disableColumnWidthUpdates(true),
-    m_time(0),
-    m_widthsDirty(true),
-    m_searchEnabled(true),
-    m_fileName(playlistFile.canonicalFilePath()),
-    m_rmbMenu(0),
-    m_toolTip(0),
-    m_blockDataChanged(false)
-{
-    setup();
-    loadFile(m_fileName, playlistFile);
-    collection->setupPlaylist(this, iconName);
-}
-
-Playlist::Playlist(PlaylistCollection *collection, bool delaySetup, int extraColumns) :
-    QTreeWidget(collection->playlistStack()),
-    m_collection(collection),
-    m_fetcher(new WebImageFetcher(this)),
-    m_allowDuplicates(true),
-    m_applySharedSettings(true),
-    m_columnWidthModeChanged(false),
-    m_disableColumnWidthUpdates(true),
-    m_time(0),
-    m_widthsDirty(true),
-    m_searchEnabled(true),
-    m_rmbMenu(0),
-    m_toolTip(0),
-    m_blockDataChanged(false)
-{
-    for(int i = 0; i < extraColumns; ++i) {
-        addColumn(i18n("JuK")); // Placeholder text!
+    // Any added columns must precede normal ones, which are normally added
+    // in setup()
+    for(int i = 0; i < extraCols; ++i) {
+        addColumn(i18n("JuK")); // Placeholder text
     }
 
     setup();
 
-    if(!delaySetup)
-        collection->setupPlaylist(this, "audio-midi");
+    // Some subclasses need to do even more handling but will remember to
+    // call setupPlaylist
+    if(!delaySetup) {
+        collection->setupPlaylist(this, iconName);
+    }
+}
+
+Playlist::Playlist(PlaylistCollection *collection, const QString &name,
+                   const QString &iconName)
+    : Playlist(false, name, collection, iconName, 0)
+{
+}
+
+Playlist::Playlist(PlaylistCollection *collection, const PlaylistItemList &items,
+                   const QString &name, const QString &iconName)
+    : Playlist(false, name, collection, iconName, 0)
+{
+    createItems(items);
+}
+
+Playlist::Playlist(PlaylistCollection *collection, const QFileInfo &playlistFile,
+                   const QString &iconName)
+    : Playlist(true, QString(), collection, iconName, 0)
+{
+    m_fileName = playlistFile.canonicalFilePath();
+    loadFile(m_fileName, playlistFile);
+    collection->setupPlaylist(this, iconName);
+}
+
+Playlist::Playlist(PlaylistCollection *collection, bool delaySetup, int extraColumns)
+    : Playlist(delaySetup, QString(), collection, QStringLiteral("audio-midi"), extraColumns)
+{
 }
 
 Playlist::~Playlist()
@@ -214,8 +179,6 @@ Playlist::~Playlist()
 
     clearItems(items());
 
-    /* delete m_toolTip; */
-
     if(!m_shuttingDown)
         m_collection->removePlaylist(this);
 }
@@ -233,31 +196,6 @@ FileHandle Playlist::currentFile() const
     return playingItem() ? playingItem()->file() : FileHandle();
 }
 
-int Playlist::time() const
-{
-    // Since this method gets a lot of traffic, let's optimize for such.
-
-    if(!m_addTime.isEmpty()) {
-        foreach(const PlaylistItem *item, m_addTime) {
-            if(item)
-                m_time += item->file().tag()->seconds();
-        }
-
-        m_addTime.clear();
-    }
-
-    if(!m_subtractTime.isEmpty()) {
-        foreach(const PlaylistItem *item, m_subtractTime) {
-            if(item)
-                m_time -= item->file().tag()->seconds();
-        }
-
-        m_subtractTime.clear();
-    }
-
-    return m_time;
-}
-
 void Playlist::playFirst()
 {
     TrackSequenceManager::instance()->setNextItem(static_cast<PlaylistItem *>(
@@ -300,10 +238,10 @@ void Playlist::playPrevious()
 
     bool random = action("randomPlay") && action<KToggleAction>("randomPlay")->isChecked();
 
-    PlaylistItem *previous = 0;
+    PlaylistItem *previous = nullptr;
 
     if(random && !m_history.isEmpty()) {
-        PlaylistItemList::Iterator last = --m_history.end();
+        PlaylistItemList::Iterator last = m_history.end() - 1;
         previous = *last;
         m_history.erase(last);
     }
@@ -369,8 +307,6 @@ void Playlist::updateDeletedItem(PlaylistItem *item)
     m_search.clearItem(item);
 
     m_history.removeAll(item);
-    m_addTime.removeAll(item);
-    m_subtractTime.removeAll(item);
 }
 
 void Playlist::clearItem(PlaylistItem *item)
@@ -780,35 +716,8 @@ void Playlist::removeFromDisk(const PlaylistItemList &items)
 void Playlist::synchronizeItemsTo(const PlaylistItemList &itemList)
 {
     // direct call to ::items to avoid infinite loop, bug 402355
-    const auto &existingItems = Playlist::items();
-    if(qAbs(itemList.count() - existingItems.count()) >
-        qMax(itemList.count(), existingItems.count()) / 2)
-    {
-        // Large imbalance in list sizes, just clear all and add without
-        // figuring out the diff also
-        clearItems(existingItems);
-        createItems(itemList);
-        return;
-    }
-
-    // Determine differences between existing playlist items and patch up
-    QHash<CollectionListItem *, PlaylistItem *> oldItems;
-    oldItems.reserve(qMax(existingItems.count(), itemList.count()));
-
-    for(const auto &item : existingItems) {
-        oldItems.insert(item->collectionItem(), item);
-    }
-
-    PlaylistItemList newItems;
-
-    for(const auto &item : itemList) {
-        if(oldItems.remove(item->collectionItem()) == 0) {
-            newItems.append(item->collectionItem());
-        }
-    }
-
-    clearItems(PlaylistItemList(oldItems.values()));
-    createItems(newItems);
+    clearItems(Playlist::items());
+    createItems(itemList);
 }
 
 void Playlist::dragEnterEvent(QDragEnterEvent *e)
@@ -887,18 +796,14 @@ bool Playlist::eventFilter(QObject *watched, QEvent *e)
 void Playlist::keyPressEvent(QKeyEvent *event)
 {
     if(event->key() == Qt::Key_Up) {
-        using ::operator|;
-        QTreeWidgetItemIterator selected(this, QTreeWidgetItemIterator::Selected |
-                                           QTreeWidgetItemIterator::NotHidden);
-        if(*selected) {
+        const auto topItem = topLevelItem(0);
+        if(topItem && topItem == currentItem()) {
             QTreeWidgetItemIterator visible(this, QTreeWidgetItemIterator::NotHidden);
-            if(*selected == *visible)
-                QApplication::postEvent(parent(), new FocusUpEvent);
+            if(topItem == *visible) {
+                emit signalMoveFocusAway();
+                event->accept();
+            }
         }
-
-    }
-    else if(event->key() == Qt::Key_Return || event->key() == Qt::Key_Enter) {
-        slotPlayCurrent();
     }
 
     QTreeWidget::keyPressEvent(event);
@@ -1098,20 +1003,11 @@ void Playlist::drawRow(QPainter *p, const QStyleOptionViewItem &option, const QM
 
 void Playlist::insertItem(QTreeWidgetItem *item)
 {
-    // Because we're called from the PlaylistItem ctor, item may not be a
-    // PlaylistItem yet (it would be QListViewItem when being inserted.  But,
-    // it will be a PlaylistItem by the time it matters, but be careful if
-    // you need to use the PlaylistItem from here.
-
-    m_addTime.append(static_cast<PlaylistItem *>(item));
     QTreeWidget::insertTopLevelItem(0, item);
 }
 
 void Playlist::takeItem(QTreeWidgetItem *item)
 {
-    // See the warning in Playlist::insertItem.
-
-    m_subtractTime.append(static_cast<PlaylistItem *>(item));
     int index = indexOfTopLevelItem(item);
     QTreeWidget::takeTopLevelItem(index);
 }
@@ -1129,31 +1025,50 @@ PlaylistItem *Playlist::createItem(const FileHandle &file, QTreeWidgetItem *afte
 
 void Playlist::createItems(const PlaylistItemList &siblings, PlaylistItem *after)
 {
-    createItems<PlaylistItem, PlaylistItem>(siblings, after);
+    createItems<QVector, PlaylistItem, PlaylistItem>(siblings, after);
 }
 
 void Playlist::addFiles(const QStringList &files, PlaylistItem *after)
 {
-    if(!after)
-        after = static_cast<PlaylistItem *>(topLevelItem(topLevelItemCount() - 1));
-
-    QApplication::setOverrideCursor(Qt::WaitCursor);
+    if(Q_UNLIKELY(files.isEmpty())) {
+        return;
+    }
 
     m_blockDataChanged = true;
+    setEnabled(false);
 
-    FileHandleList queue;
+    QVector<QFuture<void>> pendingFutures;
+    for(const auto &file : files) {
+        // some files added here will launch threads that we must wait until
+        // they're done to cleanup
+        auto pendingResult = addUntypedFile(file, after);
+        if(!pendingResult.isFinished()) {
+            pendingFutures.push_back(pendingResult);
+            ++m_itemsLoading;
+        }
+    }
 
-    foreach(const QString &file, files)
-        addFile(file, queue, true, &after);
+    // It's possible for no async threads to be launched, and also possible
+    // for this function to be called while there were other threads in flight
+    if(pendingFutures.isEmpty() && m_itemsLoading == 0) {
+        cleanupAfterAllFileLoadsCompleted();
+        return;
+    }
 
-    addFileHelper(queue, &after, true);
+    // Build handlers for all the still-active loaders on the heap and then
+    // return to the event loop.
+    for(const auto &future : qAsConst(pendingFutures)) {
+        auto loadWatcher = new QFutureWatcher<void>(this);
+        loadWatcher->setFuture(future);
 
-    m_blockDataChanged = false;
+        connect(loadWatcher, &QFutureWatcher<void>::finished, this, [=]() {
+                if(--m_itemsLoading == 0) {
+                    cleanupAfterAllFileLoadsCompleted();
+                }
 
-    slotWeightDirty();
-    playlistItemsChanged();
-
-    QApplication::restoreOverrideCursor();
+                loadWatcher->deleteLater();
+            });
+    }
 }
 
 void Playlist::refreshAlbums(const PlaylistItemList &items, coverKey id)
@@ -1444,6 +1359,7 @@ void Playlist::slotPlayFromBackMenu(QAction *backAction) const
 
 void Playlist::setup()
 {
+    setAlternatingRowColors(true);
     setRootIsDecorated(false);
     setContextMenuPolicy(Qt::CustomContextMenu);
     setUniformRowHeights(true);
@@ -1459,6 +1375,18 @@ void Playlist::setup()
 
     sortByColumn(1, Qt::AscendingOrder);
 
+    // Should this be itemActivated? It is quite annoying when I try it...
+    connect(this, &QTreeWidget::itemDoubleClicked, this, &Playlist::slotPlayCurrent);
+
+    // Use a timer to soak up the multiple dataChanged signals we're going to get
+    auto updateRequestor = new QTimer(this);
+    updateRequestor->setSingleShot(true);
+    updateRequestor->setInterval(10);
+
+    connect(model(), &QAbstractItemModel::dataChanged,
+            updateRequestor, static_cast<void (QTimer::*)()>(&QTimer::start));
+    connect(updateRequestor, &QTimer::timeout, this, &Playlist::slotUpdateTime);
+
     // This apparently must be created very early in initialization for other
     // Playlist code requiring m_headerMenu.
     m_columnVisibleAction = new KActionMenu(i18n("&Show Columns"), this);
@@ -1487,7 +1415,6 @@ void Playlist::loadFile(const QString &fileName, const QFileInfo &fileInfo)
 
     setSortingEnabled(false);
 
-    PlaylistItem *after = nullptr;
     m_disableColumnWidthUpdates = true;
     m_blockDataChanged = true;
 
@@ -1502,7 +1429,7 @@ void Playlist::loadFile(const QString &fileName, const QFileInfo &fileInfo)
         if(item.exists() && item.isFile() && item.isReadable() &&
            MediaFiles::isMediaFile(item.fileName()))
         {
-            after = createItem(FileHandle(item), after);
+            (void) createItem(FileHandle(item));
         }
     }
 
@@ -1572,8 +1499,7 @@ void Playlist::calculateColumnWeights()
     if(m_disableColumnWidthUpdates)
         return;
 
-    PlaylistItemList l = items();
-    QList<int>::Iterator columnIt;
+    const PlaylistItemList l = items();
 
     QVector<double> averageWidth(columnCount());
     double itemCount = l.size();
@@ -1611,98 +1537,95 @@ void Playlist::calculateColumnWeights()
     m_weightDirty.clear();
 }
 
-void Playlist::addFile(const QString &file, FileHandleList &files, bool importPlaylists,
-                       PlaylistItem **after)
+void Playlist::addPlaylistFile(const QString &m3uFile)
+{
+    if (!m_collection->containsPlaylistFile(m3uFile)) {
+        new Playlist(m_collection, QFileInfo(m3uFile));
+    }
+}
+
+QFuture<void> Playlist::addFilesFromDirectory(const QString &dirPath)
+{
+    auto loader = new DirectoryLoader(dirPath);
+
+    connect(loader, &DirectoryLoader::loadedPlaylist, this,
+        [this](const QString &m3uFile) {
+            addPlaylistFile(m3uFile);
+        }
+    );
+    connect(loader, &DirectoryLoader::loadedFiles, this,
+        [this](const FileHandleList &newFiles) {
+            for(const auto newFile : newFiles) {
+                createItem(newFile);
+            }
+        }
+    );
+
+    auto future = QtConcurrent::run(loader, &DirectoryLoader::startLoading);
+    auto loadWatcher = new QFutureWatcher<void>(this);
+    connect(loadWatcher, &QFutureWatcher<void>::finished, this, [=]() {
+            loader->deleteLater();
+            loadWatcher->deleteLater();
+        });
+
+    return future;
+}
+
+// Returns a future since some codepaths will result in an async operation.
+QFuture<void> Playlist::addUntypedFile(const QString &file, PlaylistItem *after)
 {
     if(hasItem(file) && !m_allowDuplicates)
-        return;
-
-    addFileHelper(files, after);
-
-    // Our biggest thing that we're fighting during startup is too many stats
-    // of files.  Make sure that we don't do one here if it's not needed.
-
-    const CollectionListItem *item = CollectionList::instance()->lookup(file);
-
-    if(item && !item->file().isNull()) {
-        FileHandle cached(item->file());
-        cached.tag();
-        files.append(cached);
-        return;
-    }
-
-    const QFileInfo fileInfo(QDir::cleanPath(file));
-    if(!fileInfo.exists())
-        return;
+        return {};
 
+    const QFileInfo fileInfo(file);
     const QString canonicalPath = fileInfo.canonicalFilePath();
 
-    if(fileInfo.isFile() && fileInfo.isReadable()) {
-        if(MediaFiles::isMediaFile(file)) {
-            FileHandle f(fileInfo);
-            f.tag();
-            files.append(f);
-        }
+    if(fileInfo.isFile() && fileInfo.isReadable() &&
+        MediaFiles::isMediaFile(file))
+    {
+        FileHandle f(fileInfo);
+        f.tag();
+        createItem(f, after);
+        return {};
     }
 
-    if(importPlaylists && MediaFiles::isPlaylistFile(file) &&
-       !m_collection->containsPlaylistFile(canonicalPath))
-    {
-        new Playlist(m_collection, fileInfo);
-        return;
+    if(MediaFiles::isPlaylistFile(file)) {
+        addPlaylistFile(canonicalPath);
+        return {};
     }
 
     if(fileInfo.isDir()) {
         foreach(const QString &directory, m_collection->excludedFolders()) {
             if(canonicalPath.startsWith(directory))
-                return; // Exclude it
+                return {}; // Exclude it
         }
 
-        QDirIterator dirIterator(canonicalPath, QDir::AllEntries | QDir::NoDotAndDotDot);
-
-        while(dirIterator.hasNext()) {
-            // We set importPlaylists to the value from the add directories
-            // dialog as we want to load all of the ones that the user has
-            // explicitly asked for, but not those that we find in lower
-            // directories.
-
-            addFile(dirIterator.next(), files,
-                    m_collection->importPlaylists(), after);
-        }
+        return addFilesFromDirectory(canonicalPath);
     }
+
+    return {};
 }
 
-void Playlist::addFileHelper(FileHandleList &files, PlaylistItem **after, bool ignoreTimer)
+// Called directly or after a threaded directory load has completed, managed by
+// m_itemsLoading
+void Playlist::cleanupAfterAllFileLoadsCompleted()
 {
-    static QTime time = QTime::currentTime();
+    m_blockDataChanged = false;
+    setEnabled(true);
 
-    // Process new items every 10 seconds, when we've loaded 1000 items, or when
-    // it's been requested in the API.
+    // Even if doing a manual column weights we'll generally start off with
+    // incorrect column sizes so at least figure out a reasonable column size
+    // and let user adjust from there.
+    if(manualResize()) {
+        auto manualResizeAction = action<KToggleAction>("resizeColumnsManually");
 
-    if(ignoreTimer || time.elapsed() > 10000 ||
-       (files.count() >= 1000 && time.elapsed() > 1000))
-    {
-        time.restart();
-
-        const bool focus = hasFocus();
-        const bool visible = isVisible() && files.count() > 20;
-
-        if(visible)
-            m_collection->raiseDistraction();
-
-        PlaylistItem *newAfter = *after;
-        foreach(const FileHandle &fileHandle, files)
-            newAfter = createItem(fileHandle, newAfter);
-        *after = newAfter;
-
-        files.clear();
-
-        if(visible)
-            m_collection->lowerDistraction();
-
-        if(focus)
-            setFocus();
+        manualResizeAction->toggle();
+        calculateColumnWeights();
+        slotUpdateColumnWidths();
+        manualResizeAction->toggle();
     }
+
+    playlistItemsChanged();
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -2103,6 +2026,20 @@ void Playlist::slotPlayCurrent()
     action("forward")->trigger();
 }
 
+void Playlist::slotUpdateTime()
+{
+    int newTime = 0;
+    QTreeWidgetItemIterator it(this);
+    while(*it) {
+        const auto item = static_cast<const PlaylistItem*>(*it);
+        ++it;
+
+        newTime += item->file().tag()->seconds();
+    }
+
+    m_time = newTime;
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 // helper functions
 ////////////////////////////////////////////////////////////////////////////////
